#include <stdlib.h>

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include "defines.h"
#include "fs.h"
#include "console.h"
#include "animation.h"

#include "player.h"
#include "particle.h"
#include "weapon.h"
#include "special.h"

/** Globals **/
SDL_Surface *screen;

int break_eventloop;

Uint32 col_gray,col_snow,col_clay;

/** External callbacks **/
void (*cntrl_keyhandler)(SDL_KeyboardEvent *key,Uint8 type);

/** Functions used only in this module **/
void con_cleanup(void);

/** Initalize SDL **/
void init_sdl() {
  char name[128];
  if(SDL_Init(SDL_INIT_VIDEO)) {
    printf("Error: Unable to initialize SDL: %s\n",SDL_GetError());
    exit(1);
  }
  sprintf(name,"%s %s",PACKAGE,VERSION);
  SDL_WM_SetCaption(name,PACKAGE);
  atexit(con_cleanup);
  break_eventloop=0;
}
/** Initialize video **/
void init_video(DisplayOptions *opts) {
  screen = SDL_SetVideoMode(SCREEN_W,SCREEN_H,SCREEN_DEPTH,opts->fullscreen?SDL_FULLSCREEN:0);
  if(screen==NULL) {
    printf("Unable to set video mode: %s\n", SDL_GetError());
    exit(1);
  }
  if(opts->hidemouse) SDL_ShowCursor(SDL_DISABLE);
  /* Initialize colours */
  col_gray=SDL_MapRGB(screen->format,128,128,128);
  col_snow=SDL_MapRGB(screen->format,176,193,255);
  col_clay=SDL_MapRGB(screen->format,255,200,128);
}

/** Cleanup **/
void con_cleanup(void) {
  deinit_players();
  deinit_particles();
  deinit_weapons(0);
  deinit_specials(0);
  SDL_Quit();
}

/** Register callbacks **/
void register_keyboard(void (*callback)(SDL_KeyboardEvent*,Uint8)) {
  cntrl_keyhandler=callback;
}

/** Game event loop **/
int event_loop() {
  SDL_Event Event;
  break_eventloop=0;
  while(1) {
    if(!SDL_WaitEvent(&Event)) {
      printf("Error occured while waiting for an event: %s\n",SDL_GetError());
      exit(1);
    }
    switch(Event.type) {
      case SDL_KEYDOWN:
      case SDL_KEYUP:
        if(Event.key.keysym.sym==SDLK_ESCAPE && Event.type==SDL_KEYDOWN) return 0;
	if(Event.key.keysym.sym==SDLK_F10 && Event.type==SDL_KEYDOWN) screenshot();
        cntrl_keyhandler(&Event.key,Event.type); break;
      case SDL_USEREVENT:
        redraw_frame();
      default:
        break;
    }
    if(break_eventloop) break;
  }
  return break_eventloop;
}

/**** Draw a box *****/
void draw_box(int x,int y,int w,int h,char width,Uint32 color) {
  SDL_Rect rect;
  rect.x=x;
  rect.y=y;
  rect.w=w;
  rect.h=width;
  SDL_FillRect(screen,&rect,color);
  rect.y+=h-width;
  SDL_FillRect(screen,&rect,color);
  rect.x=x;
  rect.y=y;
  rect.w=width;
  rect.h=h;
  SDL_FillRect(screen,&rect,color);
  rect.x+=w-width;
  SDL_FillRect(screen,&rect,color);
}

/* Draw a single pixel */
void putpixel(SDL_Surface *surface,int x,int y,Uint32 color) {
  Uint8 *bits,bpp;
  if(x>surface->w || y>surface->h) return;
  bpp = surface->format->BytesPerPixel;
  bits = ((Uint8 *)surface->pixels)+y*surface->pitch+x*bpp;
  switch(bpp) {
    case 1:
     *((Uint8 *)(bits)) = (Uint8)color;
     break;
    case 2:
      *((Uint16 *)(bits)) = (Uint16)color;
      break;
    case 3: {
      Uint8 r, g, b;
      r = (color>>surface->format->Rshift)&0xFF;
      g = (color>>surface->format->Gshift)&0xFF;
      b = (color>>surface->format->Bshift)&0xFF;
      *((bits)+surface->format->Rshift/8) = r;
      *((bits)+surface->format->Gshift/8) = g;
      *((bits)+surface->format->Bshift/8) = b;
      }
      break;
    case 4:
      *((Uint32 *)(bits)) = (Uint32)color;
  }
}

/* Modify the colors of a surface (Used to change the player colors in ship sprites) */
void recolor(SDL_Surface *surface,float red,float green,float blue,float alpha) {
  Uint8 *bits;
  int pos;
  if(surface->format->BytesPerPixel != 4) {
    printf("Error ! Unsupported color depth (%d bytes per pixel), currently only 4 bps is supported ! (fix this at recolor(), console.c)\n",screen->format->BytesPerPixel);
    exit(1);
  }
  for(pos=0;pos<surface->w*surface->h;pos++) {
     bits =((Uint8*)surface->pixels)+pos*4;
     bits[0]*=blue;
     bits[1]*=green;
     bits[2]*=red;
     bits[3]*=alpha;
  }
}

