#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <dirent.h>
#include <SDL.h>

#include "defines.h"
#include "stringutil.h"
#include "console.h"
#include "fs.h"
#include "game.h"
#include "player.h"
#include "intro.h"
#include "SFont.h"

#ifdef CLOSEDIR_VOID
/* Fake a return value. */
# define CLOSEDIR(d) (closedir (d), 0)
#else
# define CLOSEDIR(d) closedir (d)
#endif

/* Since init_intro_screen() is most likely called by now,(if not, then we are not doing things right) we can safely use these */
extern SDL_Surface *intr_logo;
extern SDL_Rect intr_logo_rect;
extern SDL_Surface *intr_arrow;
/* And some globals */
SDL_Surface *gam_filler;
GameInfo game_settings;
GameStatus game_status;
int gam_screen;
int gam_selection;

void init_game() {
  gam_filler=load_image(GFX_DIRECTORY,"filler.png",1,0);
  if(gam_filler) recolor(gam_filler,0.4,0.4,0.4,1.0); /* Think the background logo is too bright or dark ? Then adjust these values ! */
  /* Set up the default keys */
  game_settings.buttons[0][0]=SDLK_UP;
  game_settings.buttons[0][1]=SDLK_DOWN;
  game_settings.buttons[0][2]=SDLK_LEFT;
  game_settings.buttons[0][3]=SDLK_RIGHT;
  game_settings.buttons[0][4]=SDLK_RSHIFT;
  game_settings.buttons[0][5]=SDLK_RCTRL;

  game_settings.buttons[1][0]=SDLK_w;
  game_settings.buttons[1][1]=SDLK_s;
  game_settings.buttons[1][2]=SDLK_a;
  game_settings.buttons[1][3]=SDLK_d;
  game_settings.buttons[1][4]=SDLK_LSHIFT;
  game_settings.buttons[1][5]=SDLK_LCTRL;

  game_settings.buttons[2][0]=SDLK_i;
  game_settings.buttons[2][1]=SDLK_k;
  game_settings.buttons[2][2]=SDLK_j;
  game_settings.buttons[2][3]=SDLK_l;
  game_settings.buttons[2][4]=SDLK_y;
  game_settings.buttons[2][5]=SDLK_h;

  game_settings.buttons[3][0]=SDLK_KP8;
  game_settings.buttons[3][1]=SDLK_KP5;
  game_settings.buttons[3][2]=SDLK_KP4;
  game_settings.buttons[3][3]=SDLK_KP6;
  game_settings.buttons[3][4]=SDLK_KP1;
  game_settings.buttons[3][5]=SDLK_KP2;

  game_settings.ship_collisions=1;
  game_settings.indstr_base=1;
  game_settings.coll_damage=0;
  game_settings.jumpgates=0;
  game_settings.jumplife=1;
  game_settings.levels=NULL;
}

void reinit_game() {
  int p;
  for(p=0;p<4;p++) {
    game_settings.players_in[p]=' ';
    game_status.wins[p]=0;
    game_status.lifetime[p]=0;
  }
  gam_selection=0;
  game_settings.rounds=5;
  game_status.lastwin=0;
}

void prematch_game() {
  /* Randomize level */
  int r,n;
  n=rand()%game_settings.levelcount;
  game_settings.levels=game_settings.first_level;
  for(r=0;r<n;r++)
    if(game_settings.levels->next==NULL) game_settings.levels=game_settings.first_level; else game_settings.levels=game_settings.levels->next;
}

void draw_begin_game() {
  char tmps[256],p;
  gam_screen=0;
  SDL_FillRect(screen,NULL,0);
  SDL_BlitSurface(intr_logo,NULL,screen,&intr_logo_rect);
  XCenteredString2(screen,font3,MENU_ITEM1-30,"Who's in ?");
  for(p=0;p<4;p++) {
    sprintf(tmps,"%cPlayer %d",game_settings.players_in[(int)p],p+1);
    PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*p,tmps);
  }
  XCenteredString2(screen,font3,MENU_ITEM1+MENU_SPACING*4,"Number of rounds");
  sprintf(tmps,"%d (+/-)",game_settings.rounds);
  PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*5,tmps);
  XCenteredString2(screen,font3,MENU_ITEM1+MENU_SPACING*6,"Go !");
  XCenteredString2(screen,font2,MENU_ITEM1+MENU_SPACING*7,"Cancel");
  game_redraw_selection();
  SDL_UpdateRect(screen,0,0,0,0);
}

void draw_weapon_selection() {
  char tmps[256];
  Uint32 color;
  int r,l;
  SDL_Rect rect1,rect2;
  color=SDL_MapRGB(screen->format,200,80,80);
  gam_screen=1;
  /* Draw background */
  rect1.x=SCREEN_W/2-200; rect1.y=SCREEN_H/2-100;
  rect1.w=400; rect1.h=200;
  rect2.x=rect1.x+2; rect2.y=rect1.y+2;
  rect2.w=rect1.w-4; rect2.h=rect1.h-4;
  SDL_FillRect(screen,&rect1,color);
  SDL_FillRect(screen,&rect2,0);
  /* Draw texts */
  if(game_status.lastwin!=0) {
    if(game_status.lastwin<0)
      strcpy(tmps,"Tie game !");
    else
      sprintf(tmps,"Player %d won !",game_status.lastwin);
    XCenteredString2(screen,font2,rect2.y+5,tmps);
  }
  XCenteredString2(screen,font2,rect2.y+5+MENU_SPACING*2,"Next level");
  sprintf(tmps,"< %s >",game_settings.levels->name);
  PutString2(screen,font1,MENU_X,rect2.y+5+MENU_SPACING*3,tmps);
  XCenteredString2(screen,font2,rect2.y+5+MENU_SPACING*4,"Weapons");
  l=5;
  for(r=0;r<4;r++) {
    if(game_settings.players_in[r]==' ') continue;
    sprintf(tmps,"Player %d: %s",r+1,weap2str(players[r].special));
    PutString2(screen,font1,MENU_X,rect2.y+5+MENU_SPACING*l,tmps);
    l++;
  }
  /* Update screen */
  SDL_UpdateRect(screen,rect1.x,rect1.y,rect1.w,rect1.h);
}

void game_keyhandler(SDL_KeyboardEvent *event,Uint8 type) {
  int sval,p;
  if(type != SDL_KEYDOWN) return;
  switch((int)event->keysym.sym) {
    case SDLK_UP:
      if(gam_screen==1) goto goto_change_weap; 
      gam_selection--;
      if(gam_selection<0) gam_selection=7;
      if(gam_selection==4) gam_selection--;
      if(gam_selection==6 && game_settings.players_in[0]==' ' && game_settings.players_in[1]==' ' && game_settings.players_in[2]==' ' && game_settings.players_in[3]==' ')
      gam_selection--;
      game_redraw_selection();
      break;
    case SDLK_DOWN:
      if(gam_screen==1) goto goto_change_weap;
      gam_selection++;
      if(gam_selection>7) gam_selection=0;
      if(gam_selection==4) gam_selection++;
      if(gam_selection==6 && game_settings.players_in[0]==' ' && game_settings.players_in[1]==' ' && game_settings.players_in[2]==' ' && game_settings.players_in[3]==' ')
        gam_selection++;
      game_redraw_selection();
      break;
    case SDLK_RIGHT:
    case SDLK_PAGEUP:
    case SDLK_KP_PLUS:
    case SDLK_PLUS:
      if(gam_screen==0) {
        if(gam_selection==5) {
          game_settings.rounds+=(event->keysym.sym==SDLK_PAGEUP)?10:1;
          if(game_settings.rounds>=100) game_settings.rounds=1;
          draw_begin_game();
        }
      } else {
	if(game_settings.levels->next==NULL) game_settings.levels=game_settings.first_level; else game_settings.levels=game_settings.levels->next;
        draw_weapon_selection();
      }
      break;
    case SDLK_LEFT:
    case SDLK_PAGEDOWN:
    case SDLK_KP_MINUS:
    case SDLK_MINUS:
      if(gam_screen==0) {
        if(gam_selection==5) {
          game_settings.rounds-=(event->keysym.sym==SDLK_PAGEDOWN)?10:1;
          if(game_settings.rounds<=0) game_settings.rounds=99;
          draw_begin_game();
        }
      } else {
        gam_selection--;
	if(game_settings.levels->prev==NULL) game_settings.levels=game_settings.last_level; else game_settings.levels=game_settings.levels->prev;
        draw_weapon_selection();
      }
      break;
    case SDLK_KP_ENTER:
    case SDLK_RETURN:
      if(gam_screen==0) {
        sval=game_exec_selection();
        if(sval) break_eventloop=sval;
      } else {
        break_eventloop=gam_selection+1;
      }
      break;
    default:
    goto_change_weap:	/* This is one of those ultrarare times goto is actually useful ! */
      for(p=0;p<4;p++) {
        if(event->keysym.sym == game_settings.buttons[p][1] || event->keysym.sym == game_settings.buttons[p][2]) players[p].special--;
        else if(event->keysym.sym == game_settings.buttons[p][0] || event->keysym.sym == game_settings.buttons[p][3]) players[p].special++;
        if(players[p].special<1) players[p].special=WeaponCount-1;
        else if(players[p].special==WeaponCount) players[p].special=1;
      }
      draw_weapon_selection();
      break;
  }
}

void game_redraw_selection() {
  SDL_Rect rect;
  int i;
  rect.x=MENU_X-MENU_SPACING;
  rect.y=MENU_ITEM1;
  rect.w=intr_arrow->w;
  rect.h=intr_arrow->h;
  for(i=0;i<8;i++) {
    if(gam_selection==i) SDL_BlitSurface(intr_arrow,NULL,screen,&rect); else SDL_FillRect(screen,&rect,0);
    rect.y+=MENU_SPACING;
  }
  SDL_UpdateRect(screen,MENU_X-MENU_SPACING,MENU_ITEM1,rect.w,MENU_SPACING*8+MENU_SPACING);
}

void game_statistics() {
  Uint32 color;
  SDL_Event event;
  SDL_Rect rect1,rect2;
  int p,pos,winner,wins,draw[3],draws;
  char tmps[256];
  color=SDL_MapRGB(screen->format,200,80,80);
  /* Draw background */
  rect1.x=SCREEN_W/2-300; rect1.y=SCREEN_H/2-200;
  rect1.w=600; rect1.h=400;
  rect2.x=rect1.x+2; rect2.y=rect1.y+2;
  rect2.w=rect1.w-4; rect2.h=rect1.h-4;
  SDL_FillRect(screen,&rect1,color);
  SDL_FillRect(screen,&rect2,0);
  XCenteredString2(screen,font1,rect2.y+10,"Game Over");
  pos=2;
  wins=0; winner=-1; draws=0;
  PutString2(screen,font3,MENU_X,rect2.y+50+MENU_SPACING,"Player     Wins    Average lifetime");
  for(p=0;p<4;p++) {
    draw[p]=0;
    if(players_active[p]==0) continue;
    sprintf(tmps,"%d               %d            %ld s",p+1,game_status.wins[p],game_status.lifetime[p]/25/game_status.total_rounds);
    PutString2(screen,font1,MENU_X,rect2.y+50+MENU_SPACING*pos,tmps);
    if(game_status.wins[p]>wins) {wins=game_status.wins[p]; winner=p+1;}
    else if(game_status.wins[p]==wins) { draw[p]=1; draws++; }
    pos++;
  }
  pos+=2;
  if(draws && winner<0) {
    strcpy(tmps,"Draw between players ");
    if(draw[0]) sprintf(tmps,"%s 1",tmps);
    for(p=1;p<4;p++)
      if(draw[p]) sprintf(tmps,"%s, %d",tmps,p+1);
  } else sprintf(tmps,"Player %d is the winner !",winner);
  XCenteredString2(screen,font2,rect2.y+50+MENU_SPACING*pos,tmps);
  XCenteredString2(screen,font1,rect2.y+rect2.h-30,"Press enter to continue");
  SDL_UpdateRect(screen,rect1.x,rect1.y,rect1.w,rect1.h);
  /* Wait for the 'Any Key' */
  while(1) {
    SDL_WaitEvent(&event);
    if(event.type==SDL_KEYDOWN && event.key.keysym.sym==SDLK_RETURN) break;
  }
}

int game_exec_selection() {
  if(gam_selection>=0 && gam_selection<=3) {
    if(game_settings.players_in[gam_selection]==' ') game_settings.players_in[gam_selection]='+'; else game_settings.players_in[gam_selection]=' ';
    draw_begin_game();
  }
  else if(gam_selection==6) return MNU_RET_SCREEN1;
  else if(gam_selection==7) return MNU_RET_BACK;
  return 0;
}

void fill_unused_player_screens() {
  SDL_Rect rect;
  int p;
  rect.w=320;
  rect.h=320;
  rect.x=0;
  rect.y=0;
  for(p=0;p<4;p++) {
    if(gam_filler) SDL_BlitSurface(gam_filler,NULL,screen,&rect); else SDL_FillRect(screen,&rect,0);
    rect.x+=320;
    if(rect.x==640) {rect.y+=240; rect.x=0;}
  }
  SDL_UpdateRect(screen,0,0,0,0);
}

void save_game_config() {
  int p,k;
  FILE *fp;
  char *filename,tmps[256];
  filename=getfullpath(HOME_DIRECTORY,CONF_FILE);
  fp=fopen(filename,"w");
  if(!fp) {
    printf("Error ! Cannot open file \"%s\" for writing !\n",filename);
    exit(1);
  }
  /* Write keys */
  for(p=0;p<4;p++) {
    sprintf(tmps,"[keys%d]\n",p+1);
    fwrite(tmps,1,strlen(tmps),fp);
    for(k=0;k<6;k++) {
      sprintf(tmps,"%d=%d\n",k,game_settings.buttons[p][k]);
      fwrite(tmps,1,strlen(tmps),fp);
    }
  }
  /* Write settings */
  strcpy(tmps,"[settings]\n");
  fwrite(tmps,1,strlen(tmps),fp);
  sprintf(tmps,"ship_collisions=%d\n",game_settings.ship_collisions);
  fwrite(tmps,1,strlen(tmps),fp);
  sprintf(tmps,"indestructable_base=%d\n",game_settings.indstr_base);
  fwrite(tmps,1,strlen(tmps),fp);
  sprintf(tmps,"collision_damage=%d\n",game_settings.coll_damage);
  fwrite(tmps,1,strlen(tmps),fp);
  sprintf(tmps,"jumpgates=%d\n",game_settings.jumpgates);
  fwrite(tmps,1,strlen(tmps),fp);
  sprintf(tmps,"jumplife=%d\n",game_settings.jumplife);
  fwrite(tmps,1,strlen(tmps),fp);

  fclose(fp);
}

void load_game_config() {
  int readmode;
  FILE *fp;
  char *filename,tmps[256],*line=NULL,*left,*right;
  filename=getfullpath(HOME_DIRECTORY,CONF_FILE);
  fp=fopen(filename,"r");
  game_settings.levelcount=0;
  scan_levels();
  if(!fp) {
    printf("No configuration file \"%s\". Using built in defaults.\n",filename);
    return;
  }
  free(filename);
  readmode=-1;
  while(fgets(tmps,255,fp)) {
    free(line);
    line=strip_white_space(tmps);
    if(strlen(line)==0) continue;
    if(line[0]=='[') readmode=-1;
    if(strcmp(line,"[keys1]")==0) {readmode=0; continue;}
    else if(strcmp(line,"[keys2]")==0) {readmode=1; continue;}
    else if(strcmp(line,"[keys3]")==0) {readmode=2; continue;}
    else if(strcmp(line,"[keys4]")==0) {readmode=3; continue;}
    else if(strcmp(line,"[settings]")==0) {readmode=4; continue;}

    if(readmode==0) {
      split_string(line,'=',&left,&right);
      game_settings.buttons[0][atoi(left)]=atoi(right);
    }
    else if(readmode==1) {
      split_string(line,'=',&left,&right);
      game_settings.buttons[1][atoi(left)]=atoi(right);
    }
    else if(readmode==2) {
      split_string(line,'=',&left,&right);
      game_settings.buttons[2][atoi(left)]=atoi(right);
    }
    else if(readmode==3) {
      split_string(line,'=',&left,&right);
      game_settings.buttons[3][atoi(left)]=atoi(right);
    }
    else if(readmode==4) {
      split_string(line,'=',&left,&right);
      if(strcmp(left,"ship_collisions")==0) game_settings.ship_collisions=atoi(right);
      else if(strcmp(left,"indestructable_base")==0) game_settings.indstr_base=atoi(right);
      else if(strcmp(left,"collision_damage")==0) game_settings.coll_damage=atoi(right);
      else if(strcmp(left,"jumpgates")==0) game_settings.jumpgates=atoi(right);
      else if(strcmp(left,"jumplife")==0) game_settings.jumplife=atoi(right);
    }
  }
  fclose(fp);
}

void scan_levels() {
  DIR *reading;
  struct dirent *next;
  char *dirname;
  char *filename;
  struct LevelFile *list=NULL;
  struct LevelFile *newentry;
  char found;
  dirname=getfullpath(LEVEL_DIRECTORY,"");
  reading=opendir(dirname);
  if(!reading) {
    printf("Error ! Cannot open directory \"%s\" !\n(Check fs.c that you have correct directory definitions)\n",dirname);
    exit(1);
  }
  while((next=readdir(reading))!=NULL) {
    filename=strdup(next->d_name);
    if(strstr(filename,".lev.")) {
      list=game_settings.levels;
      found=0;
      if(list)
	while(list) {
	  if(list->collmap) if(strncmp(list->collmap,filename,strstr(filename,".lev.")-filename)==0) {
	    list->filename=filename;
	    list->name=level_name(filename);
	    game_settings.levelcount++;
	    found=1;
	    break;
	  }
	  list=list->next;
        }
      if(found==0) {
	if(game_settings.levels!=NULL) {
	  list=game_settings.levels;
	  while(list->next) list=list->next;
	}
        newentry=(struct LevelFile*)malloc(sizeof(struct LevelFile));
        newentry->prev=list;
        newentry->next=NULL;
        newentry->filename=filename;
	newentry->collmap=NULL;
	newentry->name=level_name(filename);
        if(game_settings.levels==NULL) {
  	  game_settings.levels=newentry;
	  game_settings.first_level=newentry;
	} else {
	  list->next=newentry;
	  game_settings.last_level=newentry;
	}
      }
    } else if(strstr(filename,".coll.")) {
      list=game_settings.levels;
      found=0;
      if(list)
        while(list) {
          if(list->filename) if(strncmp(list->filename,filename,strstr(filename,".coll.")-filename)==0) {
   	    list->collmap=filename;
	    found=1;
	    game_settings.levelcount++;
	    break;
	  }
          list=list->next;
        }
      if(found==0) {
	if(game_settings.levels!=NULL) {
	  list=game_settings.levels;
	  while(list->next) list=list->next;
	}
	newentry=(struct LevelFile*)malloc(sizeof(struct LevelFile));
	newentry->prev=list;
	newentry->next=NULL;
	newentry->filename=NULL;
	newentry->collmap=filename;
	newentry->name=NULL;
	if(game_settings.levels==NULL) {
	  game_settings.levels=newentry;
	  game_settings.first_level=newentry;
	} else {
	  list->next=newentry;
	  game_settings.last_level=newentry;
	}
      }
    } else free(filename);
  }
  CLOSEDIR(reading);
}

char *level_name(char *fname) {
  int len,r;
  char *newname;
  len=strstr(fname,".lev.")-fname;
  newname=malloc(len);
  for(r=0;r<len;r++)
    newname[r]=(fname[r]!='_')?fname[r]:' ';
  newname[len]=0;
  return newname;
}
