#include <string.h>
#include <stdlib.h>
#include "SDL.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include "defines.h"
#include "fs.h"
#include "stringutil.h"
#include "console.h"
#include "SFont.h"
#include "intro.h"
#include "game.h"

/* Fonts */
SFont_FontInfo *font1,*font2,*font3;

/* Internally used globals */
SDL_Surface *intr_logo;
SDL_Surface *intr_arrow;
SDL_Rect intr_logo_rect;
int intr_selection;
int intr_screen;
int intr_screen_items;
int intr_screen_param;
int intr_screen_param2;

void init_intro_screen() {
  font1=malloc(sizeof(SFont_FontInfo));
  font2=malloc(sizeof(SFont_FontInfo));
  font3=malloc(sizeof(SFont_FontInfo));
  intr_logo=load_image(GFX_DIRECTORY,"logo.jpg",1,0);
  intr_arrow=load_image(GFX_DIRECTORY,"select.png",0,0);
  font1->Surface=load_image(GFX_DIRECTORY,"font1.png",0,1);
  font2->Surface=load_image(GFX_DIRECTORY,"font2.png",1,1);
  if(!font2->Surface) font2->Surface=font1->Surface;
  font3->Surface=load_image(GFX_DIRECTORY,"font3.png",1,1);
  if(!font3->Surface) font3->Surface=font1->Surface;
  if(intr_logo) {
    intr_logo_rect.x=SCREEN_W/2-intr_logo->w/2;
    intr_logo_rect.y=10;
  }
  InitFont2(font1);
  InitFont2(font2);
  InitFont2(font3);
  intr_selection=0;
  intr_screen_items=3;
  intr_screen=0;
  intr_screen_param2=-1;
}

void draw_intro_screen() {
  char tmps[256];
  SDL_FillRect(screen,NULL,0);
  SDL_BlitSurface(intr_logo,NULL,screen,&intr_logo_rect);
  XCenteredString2(screen,font2,MENU_ITEM1-50,VERSION);
  switch(intr_screen) {
    case SCREEN_INTRO:
      PutString2(screen,font1,MENU_X,MENU_ITEM1,"Start a new game");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING,"Configuration");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*2,"Exit");
      break;
    case SCREEN_SETTINGS:
      PutString2(screen,font1,MENU_X,MENU_ITEM1,"Player 1 keys");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING,"Player 2 keys");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*2,"Player 3 keys");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*3,"Player 4 keys");
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*4,"Game settings");
      PutString2(screen,font3,MENU_X,MENU_ITEM1+MENU_SPACING*5,"Return");
      PutString2(screen,font2,MENU_X,MENU_ITEM1+MENU_SPACING*6,"Save configuration to ~/.luola.conf");
      break;
    case SCREEN_KEYCONF:
      sprintf(tmps,"Player %d key configuration",intr_screen_param+1);
      XCenteredString2(screen,font1,MENU_ITEM1-30,tmps);
      sprintf(tmps,"Thrust - %s",key2str(game_settings.buttons[intr_screen_param][0]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1,tmps);
      sprintf(tmps,"Reverse - %s",key2str(game_settings.buttons[intr_screen_param][1]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING,tmps);
      sprintf(tmps,"Left - %s",key2str(game_settings.buttons[intr_screen_param][2]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*2,tmps);
      sprintf(tmps,"Right - %s",key2str(game_settings.buttons[intr_screen_param][3]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*3,tmps);
      sprintf(tmps,"Fire normal - %s",key2str(game_settings.buttons[intr_screen_param][4]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*4,tmps);
      sprintf(tmps,"Fire special - %s",key2str(game_settings.buttons[intr_screen_param][5]));
      PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*5,tmps);
      PutString2(screen,font3,MENU_X,MENU_ITEM1+MENU_SPACING*6,"  Ok");
      break;
    case SCREEN_GAMESETTINGS:
       XCenteredString2(screen,font2,MENU_ITEM1-30,"General game settings");
       sprintf(tmps,"Ship collisions %s",(game_settings.ship_collisions)?"enabled":"disabled");
       PutString2(screen,font1,MENU_X,MENU_ITEM1,tmps);
       sprintf(tmps,"%s bases",(game_settings.indstr_base)?"Indestructable":"Destructable");
       PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*1,tmps);
       sprintf(tmps,"%sollision damage",(game_settings.coll_damage)?"C":"No c");
       PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*2,tmps);
       sprintf(tmps,"Jump gates: %d pairs",game_settings.jumpgates);
       PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*3,tmps);
       sprintf(tmps,"Jump-point life: %s",(game_settings.jumplife)?"long":"short");
       PutString2(screen,font1,MENU_X,MENU_ITEM1+MENU_SPACING*4,tmps);
       PutString2(screen,font3,MENU_X,MENU_ITEM1+MENU_SPACING*5,"Ok");
  }
  SDL_UpdateRect(screen,0,0,0,0);
  redraw_selection();
}

void redraw_selection() {
  SDL_Rect rect;
  int i;
  rect.x=MENU_X-MENU_SPACING;
  rect.y=MENU_ITEM1;
  rect.w=intr_arrow->w;
  rect.h=intr_arrow->h;
  for(i=0;i<intr_screen_items;i++) {
    if(intr_selection==i) SDL_BlitSurface(intr_arrow,NULL,screen,&rect); else SDL_FillRect(screen,&rect,0);
    rect.y+=MENU_SPACING;
  }
  SDL_UpdateRect(screen,MENU_X-MENU_SPACING,MENU_ITEM1,rect.w,MENU_SPACING*intr_screen_items+MENU_SPACING);
}

void intro_keyhandler(SDL_KeyboardEvent *event,Uint8 type) {
  int sval;
  if(type!=SDL_KEYDOWN) return;
  if(intr_screen_param2!=-1) {
    game_settings.buttons[intr_screen_param][intr_screen_param2]=event->keysym.sym;
    draw_intro_screen();
    intr_screen_param2=-1;
    return;
  }
  switch((int)event->keysym.sym) {
    case SDLK_UP:
      intr_selection--;
      if(intr_selection<0)  intr_selection=intr_screen_items-1;
      redraw_selection();
      break;
    case SDLK_DOWN:
      intr_selection++;
      if(intr_selection>=intr_screen_items) intr_selection=0;
      redraw_selection();
      break;
    case SDLK_RETURN:
      sval=exec_selection();
      if(sval==MNU_RET_BACK) {
        if(intr_screen==SCREEN_INTRO) break_eventloop=sval;
        else if(intr_screen==SCREEN_SETTINGS) switch_to_screen(SCREEN_INTRO,0);
        else if(intr_screen==SCREEN_KEYCONF) switch_to_screen(SCREEN_SETTINGS,0);
	else if(intr_screen==SCREEN_GAMESETTINGS) switch_to_screen(SCREEN_SETTINGS,0);
      }
      if(sval==MNU_RET_SCREEN1) break_eventloop=sval;
      break;
  }
  return;
}

int exec_selection() {
  switch(intr_screen) {
    case SCREEN_INTRO:
      if(intr_selection==0) return MNU_RET_SCREEN1; /* Go to game */
      else if(intr_selection==1) switch_to_screen(SCREEN_SETTINGS,0);
      else if(intr_selection==2) return MNU_RET_BACK; /* Exit */
      break;
    case SCREEN_SETTINGS:
      if(intr_selection>=0 && intr_selection <= 3) {switch_to_screen(SCREEN_KEYCONF,intr_selection); return 0; }
      else if(intr_selection==4) {switch_to_screen(SCREEN_GAMESETTINGS,0); return 0; }
      else if(intr_selection==5) return MNU_RET_BACK;
      else if(intr_selection==6) save_game_config();
      break;
    case SCREEN_KEYCONF:
      if(intr_selection>=0 && intr_selection<=5) ask_key(intr_selection);
      if(intr_selection==6) return MNU_RET_BACK;
      break;
    case SCREEN_GAMESETTINGS:
      if(intr_selection==0) game_settings.ship_collisions=!game_settings.ship_collisions;
      else if(intr_selection==1) game_settings.indstr_base=!game_settings.indstr_base;
      else if(intr_selection==2) game_settings.coll_damage=!game_settings.coll_damage;
      else if(intr_selection==3) {game_settings.jumpgates++; if(game_settings.jumpgates>5) game_settings.jumpgates=0; }
      else if(intr_selection==4) game_settings.jumplife=!game_settings.jumplife;
      else if(intr_selection==5) return MNU_RET_BACK;
      draw_intro_screen();
      break;
  }
  return 0;
}

void switch_to_screen(int newscr,int param) {
  intr_screen=newscr;
  intr_selection=0;
  intr_screen_param=param;
  if(newscr==SCREEN_INTRO) intr_screen_items=3;
  else if(newscr==SCREEN_SETTINGS) intr_screen_items=7;
  else if(newscr==SCREEN_KEYCONF) intr_screen_items=7;
  else if(newscr==SCREEN_GAMESETTINGS) intr_screen_items=6;
  draw_intro_screen();
}

void ask_key(int k) {
  Uint32 color;
  SDL_Rect rect1,rect2;
  color=SDL_MapRGB(screen->format,200,80,80);
  intr_screen_param2=k;
  rect1.x=SCREEN_W/2-100;
  rect1.y=SCREEN_H/2-25;
  rect1.w=200;
  rect1.h=50;
  rect2.x=rect1.x+2;
  rect2.y=rect1.y+2;
  rect2.w=rect1.w-4;
  rect2.h=rect1.h-4;
  SDL_FillRect(screen,&rect1,color);
  SDL_FillRect(screen,&rect2,0);
  XCenteredString(screen,SCREEN_H/2-10,"Press a key");
  SDL_UpdateRect(screen,rect1.x,rect1.y,rect1.w,rect1.h);
}
