#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "defines.h"
#include "console.h"
#include "SDL_image.h"
#include "stringutil.h"
#include "player.h"
#include "fs.h"
#include "level.h"
#include "particle.h"

/* Level effects */
typedef struct {
  int x,y;
  int brake;
  unsigned char value;
  LevelFXType type;
} LevelFX;

/* Linked lists for level effects */
struct LevelEffects {
  LevelFX *fx;
  struct LevelEffects *next;
  struct LevelEffects *prev;
} *level_effects;

/* Globals */
SDL_Rect lev_p1_rect,lev_p2_rect,lev_p3_rect,lev_p4_rect;
Uint32 burncolor[FIRE_FRAMES];
/* Exported globals */
Level lev_level;
SDL_Rect cam1_rect,cam2_rect,cam3_rect,cam4_rect;
Vector gravity;
Uint32 lev_watercol;

void init_level() {
  int r,red,green,blue;
  gravity = makeVector(0,-GRAVITY);
  cam1_rect.w=320;
  cam1_rect.h=240;
  cam2_rect.w=320;
  cam2_rect.h=240;
  cam3_rect.w=320;
  cam3_rect.h=240;
  cam4_rect.w=320;
  cam4_rect.h=240;
  lev_p1_rect.x=0;
  lev_p1_rect.y=0;
  lev_p2_rect.x=320;
  lev_p2_rect.y=0;
  lev_p3_rect.x=0;
  lev_p3_rect.y=240;
  lev_p4_rect.x=320;
  lev_p4_rect.y=240;
  lev_watercol=SDL_MapRGB(screen->format,100,100,255);
  level_effects=NULL;
  for(r=0;r<FIRE_FRAMES;r++) {
    if(r==0) red=0; else red=128+(255.0/FIRE_FRAMES)*r;
    green=(255.0/FIRE_FRAMES)*r;
    blue=0;
    if(red>255) red=255;
    burncolor[r]=SDL_MapRGB(screen->format,red,green,blue);
  }
}

void load_level(struct LevelFile *lev) {
  SDL_Surface *collmap;
  int x,y,p;
  SDL_Color *tmpcol;
  Uint8 pixelval,*bits;
  /*
  lev_level.player_def_x[0]=100; lev_level.player_def_y[0]=100;
  lev_level.player_def_x[1]=200; lev_level.player_def_y[1]=100;
  lev_level.player_def_x[2]=300; lev_level.player_def_y[2]=100;
  lev_level.player_def_x[3]=400; lev_level.player_def_y[3]=100;
  */
  lev_level.terrain=load_image(LEVEL_DIRECTORY,lev->filename,0,0);
  lev_level.width=lev_level.terrain->w;
  lev_level.height=lev_level.terrain->h;
  collmap=IMG_Load(getfullpath(LEVEL_DIRECTORY,lev->collmap));
  if(collmap->format->BitsPerPixel!=8) {
    printf("Error ! Collision map image \"%s\" has %d bits per pixel, should be 8 !\n",lev->collmap,collmap->format->BitsPerPixel);
    exit(1);
  }
  if(collmap->w != lev_level.width || collmap->h != lev_level.height) {
    printf("Error Collision map image \"%s\" has incorrect size (%dx%d), should be %dx%d !\n",lev->collmap,collmap->w,collmap->h,lev_level.width,lev_level.height);
    exit(1);
  }
  /* Get palette entries */
  tmpcol=&collmap->format->palette->colors[TER_WATER];
  lev_watercol=SDL_MapRGB(screen->format,tmpcol->r,tmpcol->g,tmpcol->b);
  /* Load data map */
  lev_level.solid=(char**)malloc(sizeof(char*)*lev_level.width);
  for(x=0;x<lev_level.width;x++) {
    lev_level.solid[x]=(char*)malloc(sizeof(char)*lev_level.height);
    for(y=0;y<lev_level.height;y++) {
      bits = ((Uint8 *)collmap->pixels)+y*collmap->pitch+x;
      (Uint8)pixelval = *((Uint8 *)(bits));
      if(pixelval>TER_COMBUSTABLE)
        lev_level.solid[x][y]=TER_FREE;
      else
        lev_level.solid[x][y]=pixelval;
    }
  }
  /* Position players */
  for(p=0;p<4;p++)
    if(game_settings.players_in[p]!=' ') {
      do {
        lev_level.player_def_x[p]=rand()%lev_level.width;
        lev_level.player_def_y[p]=rand()%lev_level.height;
      } while(lev_level.solid[lev_level.player_def_x[p]][lev_level.player_def_y[p]]!=TER_FREE);
    }
  SDL_FreeSurface(collmap);
}

void unload_level() {
  int x;
  struct LevelEffects *next;
  SDL_FreeSurface(lev_level.terrain);
  for(x=0;x<lev_level.width;x++)
    free(lev_level.solid[x]);
  free(lev_level.solid);
  while(level_effects) {
    next=level_effects->next;
    free(level_effects->fx);
    free(level_effects);
    level_effects=next;
  }
}

void draw_level() {
  if(players_active[0]) SDL_BlitSurface(lev_level.terrain,&cam1_rect,screen,&lev_p1_rect);
  if(players_active[1]) SDL_BlitSurface(lev_level.terrain,&cam2_rect,screen,&lev_p2_rect);
  if(players_active[2]) SDL_BlitSurface(lev_level.terrain,&cam3_rect,screen,&lev_p3_rect);
  if(players_active[3]) SDL_BlitSurface(lev_level.terrain,&cam4_rect,screen,&lev_p4_rect);
}

inline signed char hit_solid(int x,int y) {
  if(lev_level.solid[x][y]==TER_FREE) return 0;
  if(lev_level.solid[x][y]==TER_WATER) return -1;
  if(lev_level.solid[x][y]>=TER_WATERFU && lev_level.solid[x][y] <= TER_WATERFL) return -2-(lev_level.solid[x][y]-TER_WATERFU);
  return lev_level.solid[x][y];
}

void start_burning(int x,int y) {
  struct LevelEffects *list=NULL;
  struct LevelEffects *newentry;
  LevelFX *fx;
  if(level_effects!=NULL) {
    list=level_effects;
    while(list->next) list=list->next;
  }
  newentry=(struct LevelEffects*)malloc(sizeof(struct LevelEffects));
  newentry->next=NULL;
  newentry->prev=list;
  fx=(LevelFX*)malloc(sizeof(LevelFX));
  fx->x=x; fx->y=y;
  fx->type=Fire;
  fx->value=FIRE_FRAMES;
  lev_level.solid[x][y]=TER_FREE;
  newentry->fx=fx;
  if(level_effects==NULL)
    level_effects=newentry;
  else
    list->next=newentry;
}

void alter_level(int x,int y,int recurse,LevelFXType type) {
  struct LevelEffects *list=NULL;
  struct LevelEffects *newentry;
  LevelFX *fx;
  if(level_effects!=NULL) {
    list=level_effects;
    while(list->next) list=list->next;
  }
  newentry=(struct LevelEffects*)malloc(sizeof(struct LevelEffects));
  newentry->next=NULL;
  newentry->prev=list;
  fx=(LevelFX*)malloc(sizeof(LevelFX));
  fx->x=x; fx->y=y;
  fx->value=3;
  fx->type=type;
  fx->brake=recurse;
  if(type==Explosive) {
    lev_level.solid[x][y]=TER_EXPLOSIVE;
  } else {
    if(lev_level.solid[x][y]==TER_FREE) lev_level.solid[x][y]=TER_GROUND;
    else if(lev_level.solid[x][y]==TER_WATER) lev_level.solid[x][y]=TER_UNDERWATER;
  }
  newentry->fx=fx;
  if(level_effects==NULL)
    level_effects=newentry;
  else
    list->next=newentry;
}

void animate_level() {
  struct LevelEffects *list=level_effects;
  struct LevelEffects *next=NULL;
  float f;
  int v;
  char solid;
  Particle *part;
  while(list) {
    next=list->next;
    if(list->fx->type==Fire) {
    list->fx->value--;
    if(list->fx->value>0)
      v=list->fx->value+rand()%FIRE_RANDOM;
    else
      v=list->fx->value;
    if(list->fx->x<3 || list->fx->y<3 || list->fx->x>lev_level.width-3 || list->fx->y>lev_level.height-3) {
	list->fx->value=0;
    }else {
    if(lev_level.solid[list->fx->x][list->fx->y+1]==TER_FREE && list->fx->value<FIRE_SPREAD) {
      putpixel(lev_level.terrain,list->fx->x,list->fx->y,burncolor[0]);
      list->fx->y-=rand()%3;
    }
    if(v<FIRE_FRAMES) 
      putpixel(lev_level.terrain,list->fx->x,list->fx->y,burncolor[v]);
    if(list->fx->value==FIRE_SPREAD) {
      for(f=-M_PI;f<M_PI;f+=M_PI/4) {
	solid=lev_level.solid[(int)(list->fx->x+sin(f)*4.0)][(int)(list->fx->y+cos(f)*4.0)];
        if(solid==TER_COMBUSTABLE) start_burning((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0));
	else if(solid==TER_EXPLOSIVE) spawn_clusters((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0),6,-1);
	else if(solid==TER_FREE && cos(f)<0 && rand()%2==1) {
	  part=make_particle((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0),9);
	  part->vector.y=2.5;
	  part->vector.x=-rand()%4;
	  part->color[0]=255;part->color[1]=178;part->color[2]=0;
	  part->targ_color[0]=100;part->targ_color[1]=100;part->targ_color[2]=100;
	  calc_color_deltas(part);
	  addParticle(part);
	}
      }
    }
    }
    } else { /* Ice, Earth or Explosive*/
      list->fx->value--;
      if(list->fx->value==1) {
	putpixel(lev_level.terrain,list->fx->x,list->fx->y,(list->fx->type==Ice)?col_snow:(list->fx->type==Earth)?col_clay:col_gray);
	if(list->fx->brake>0) list->fx->brake--;
	if(list->fx->x>5 && list->fx->y>5 && list->fx->x<lev_level.width-5 && list->fx->y<lev_level.height-5) {
	for(f=-M_PI;f<M_PI;f+=M_PI/4) {
	  solid=lev_level.solid[(int)(list->fx->x+sin(f)*4.0)][(int)(list->fx->y+cos(f)*4.0)];
	  if(list->fx->type==Ice) {
	    if(solid==TER_WATER && list->fx->brake) alter_level((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0),list->fx->brake,list->fx->type);
	  } else if(list->fx->type==Earth) {
	    if((solid==TER_FREE || solid==TER_WATER) && list->fx->brake) alter_level((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0),list->fx->brake,list->fx->type);
	  } else {
	    if((solid==TER_GROUND || solid==TER_COMBUSTABLE) && list->fx->brake) alter_level((int)(list->fx->x+sin(f)*4.0),(int)(list->fx->y+cos(f)*4.0),list->fx->brake,list->fx->type);
	  }
	}
	}
      }
    }
    if(list->fx->value==0) {
      free(list->fx);
      if(list->prev) list->prev->next=list->next;
      else level_effects=list->next;
      if(list->next)
        list->next->prev=list->prev;
      free(list);
    }
    list=next;
  }
}
