#include <stdlib.h>
#include "SDL.h"

#include "console.h"
#include "level.h"
#include "player.h"
#include "particle.h"

/* Linked list containg all particles */
struct Particle_list {
  Particle *particle;
  struct Particle_list *next;
  struct Particle_list *prev;
};

/* Internally used globals */
struct Particle_list *particles;

inline void calc_color_deltas(Particle *part);

/* Initialize */
void init_particles() {
  particles=NULL;
}

void deinit_particles() {
  struct Particle_list *next;
  while(particles) {
    next=particles->next;
    free(particles);
    particles=next;
  }
}

/* Utility function to create a new particle */
Particle *make_particle(int x,int y,int age) {
  Particle *newpart;
  newpart=(Particle*)malloc(sizeof(Particle));
  newpart->x=x;
  newpart->y=y;
  newpart->age=age;
  newpart->color[0]=200;
  newpart->color[1]=200;
  newpart->color[2]=255;
  newpart->targ_color[0]=0;
  newpart->targ_color[1]=0;
  newpart->targ_color[2]=0;
  calc_color_deltas(newpart);
  return newpart;
}

/* Add a new particle to list*/
void addParticle(Particle *newpart) {
  struct Particle_list *list=NULL;
  struct Particle_list *newentry;
  if(particles!=NULL) {
    list=particles;
    while(list->next) list=list->next;
  }
  newentry=(struct Particle_list*)malloc(sizeof(struct Particle_list));
  newentry->next=NULL;
  newentry->prev=list;
  newentry->particle=newpart;
  if(particles==NULL)
    particles=newentry;
  else
    list->next=newentry;
}

void animate_particles() {
  struct Particle_list *list=particles,*next;
  Particle *part;
  if(list==NULL) return;
  while(list) {
    part=list->particle;
    next=list->next;
    part->x-=part->vector.x;
    part->y-=part->vector.y;
    part->color[0]+=part->rd;
    part->color[1]+=part->gd;
    part->color[2]+=part->bd;
    part->age--;
    if(part->age<=0) { /* Particle has expired, delete it */
      free(part);
      if(list->prev)
        list->prev->next = list->next;
      else
        particles=list->next;
      if(list->next)
        list->next->prev = list->prev;
      free(list);
    }
    list=next;
  }
}

void draw_particles() {
  struct Particle_list *list=particles;
  Uint32 color;
  Particle *part;
  int x,y;
  if(list==NULL) return;
  if( SDL_MUSTLOCK(screen) ) SDL_LockSurface(screen);

  while(list) {
    part=list->particle;
    /*
    part->color[0]+=part->rd;
    part->color[1]+=part->gd;
    part->color[2]+=part->bd;
    */
    color=SDL_MapRGB(screen->format,part->color[0],part->color[1],part->color[2]);
    if(players_active[0]) {
      x=part->x-cam1_rect.x;
      y=part->y-cam1_rect.y;
      if((x>0 && x<320) && (y>0 && y<240))
        putpixel(screen,x,y,color);
    }
    if(players_active[1]) {
      x=part->x-cam2_rect.x;
      y=part->y-cam2_rect.y;
      if((x>0 && x<320) && (y> 0 && y<240))
        putpixel(screen,320+x,y,color);
    }
    if(players_active[2]) {
      x=part->x-cam3_rect.x;
      y=part->y-cam3_rect.y;
      if((x>0 && x<320) && (y>0 && y<240))
        putpixel(screen,x,240+y,color);
    }
    if(players_active[3]) {
      x=part->x-cam4_rect.x;
      y=part->y-cam4_rect.y;
      if((x>0 && x<320) && (y>0 && y<240))
        putpixel(screen,320+x,240+y,color);
    }
    list=list->next;
  }
  if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
}

/* Calculate color delta values */
inline void calc_color_deltas(Particle *part) {
  part->rd=(part->targ_color[0]-part->color[0])/part->age;
  part->gd=(part->targ_color[1]-part->color[1])/part->age;
  part->bd=(part->targ_color[2]-part->color[2])/part->age;
}
