#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "SDL.h"
#include "defines.h"
#include "fs.h"
#include "console.h"
#include "particle.h"
#include "animation.h"
#include "level.h"
#include "weapon.h"
#include "player.h"
#include "game.h"
#include "special.h"

/* Globals */
SDL_Surface **plr_greyship;
SDL_Surface **plr_frozenship;
Player players[4];
char players_active[4];
int plr_alive;
Uint32 plr_healthbar_col,plr_energybar_col;
int endgame;

/* Just borrowing these... */
extern Projectile *projectiles;
extern GameStatus game_status;

/* Internally used functions */
void init_player_images(int plrn);

void init_players() {
  int p;
  init_player_images(-1);
  init_player_images(-2);
  init_player_images(0);
  init_player_images(1);
  init_player_images(2);
  init_player_images(3);
  plr_healthbar_col=SDL_MapRGB(screen->format,70,88,176);
  plr_energybar_col=SDL_MapRGB(screen->format,176,185,185);
  for(p=0;p<4;p++) {
    players[p].special=WGrenade;
  }
}

void deinit_players() {
  int r;
  for(r=0;r<SHIP_POSES;r++) {
    SDL_FreeSurface(plr_greyship[r]);
    SDL_FreeSurface(plr_frozenship[r]);
    SDL_FreeSurface(players[0].ship[r]);
    SDL_FreeSurface(players[1].ship[r]);
    SDL_FreeSurface(players[2].ship[r]);
    SDL_FreeSurface(players[3].ship[r]);
  }
  free(plr_greyship);
  free(plr_frozenship);
  free(players[0].ship);
  free(players[1].ship);
  free(players[2].ship);
  free(players[3].ship);
}

void reinit_players() {
  int p;
  plr_alive=0;
  endgame=-1;
  for(p=0;p<4;p++) {
    players_active[p]=(game_settings.players_in[p]==' ')?0:1;
    players[p].thrust=0;
    players[p].turn=0;
    players[p].angle=0;
    players[p].health=1;
    players[p].energy=1;
    players[p].dead=0;
    players[p].visible=1;
    players[p].shieldup=0;
    players[p].ghost=0;
    players[p].afterburn=1;
    players[p].x=lev_level.player_def_x[p];
    players[p].y=lev_level.player_def_y[p];
    players[p].vector.x=0;
    players[p].vector.y=0;
    if(players_active[p]) plr_alive++;
  }
  if(plr_alive==1) plr_alive++;
  reinit_animation();
}

void init_player_images(int plrn) {
  int r,n,d;
  char filename[]="vwing00.tga";
  if(plrn==-1)
    plr_greyship=(SDL_Surface**)malloc(sizeof(SDL_Surface)*SHIP_POSES);
  else if(plrn==-2)
    plr_frozenship=(SDL_Surface**)malloc(sizeof(SDL_Surface)*SHIP_POSES);
  else {
    players[plrn].ship=(SDL_Surface**)malloc(sizeof(SDL_Surface)*SHIP_POSES);
    players[plrn].shield=load_image(GFX_DIRECTORY,"shield.png",0,1);
    switch(plrn) {
      case 0:
        recolor(players[plrn].shield,0.4,0.4,1,1);
        break;
      case 1:
        recolor(players[plrn].shield,1,0.4,0.4,1);
        break;
      case 2:
        recolor(players[plrn].shield,0.4,1,0.4,1);
        break;
      case 3:
        recolor(players[plrn].shield,1,1,0.6,1);
        break;
    }
  }
  for(r=0;r<SHIP_POSES;r++) {
    memset(filename+5,'0',2);
    n=r+2;
    for(d=2;d>0;d--) {
      for(;n>pow(10,d-1);) 
        if(n-=pow(10,d-1)) { filename[7-d]++; } else break;
    }
    if(plrn==-1)
      plr_greyship[r]=load_image(GFX_DIRECTORY,filename,0,1);
    else if(plrn==-2) {
      plr_frozenship[r]=load_image(GFX_DIRECTORY,filename,0,1);
      recolor(plr_frozenship[r],0,1,1,1);
    } else {
      players[plrn].ship[r]=load_image(GFX_DIRECTORY,filename,0,1);
      switch(plrn) {
        case 0:
          recolor(players[plrn].ship[r],0,0.2,1,1);
          break;
        case 1:
         recolor(players[plrn].ship[r],1,0.2,0,1);
         break;
        case 2:
          recolor(players[plrn].ship[r],0.2,1,0.2,1);
          break;
        case 3:
          recolor(players[plrn].ship[r],1,1,0.4,1);
          break;
      }
    }
  }
}

void draw_player() {
  SDL_Rect rect;
  int p,which;
  for(p=0;p<4;p++) 
    if(players_active[p] && players[p].visible) {
      for(which=0;which<4;which++) {
	if(players_active[which]) {
	  switch(which) {
	    case 0:
              rect.x=players[p].x-cam1_rect.x-8;
              rect.y=players[p].y-cam1_rect.y-8;
              if(rect.x<0 || rect.y<0 || rect.x>320 || rect.y>240) continue;
              break;
            case 1:
              rect.x=320+players[p].x-cam2_rect.x-8;
              rect.y=players[p].y-cam2_rect.y-8;
              if(rect.x<320 || rect.y>240) continue;
              break;
            case 2:
              rect.x=players[p].x-cam3_rect.x-8;
              rect.y=240+players[p].y-cam3_rect.y-8;
              if(rect.x>320 || rect.y<240) continue;
              break;
            case 3:
              rect.x=320+players[p].x-cam4_rect.x-8;
              rect.y=240+players[p].y-cam4_rect.y-8;
              if(rect.x<320 || rect.y<240) continue;
              break;
          }
        if(players[p].dead)
          SDL_BlitSurface(plr_greyship[(int)(15-players[p].angle/(2*M_PI/15.0))],NULL,screen,&rect);
        else if(players[p].frozen)
	  SDL_BlitSurface(plr_frozenship[(int)(15-players[p].angle/(2*M_PI/15.0))],NULL,screen,&rect);
	else
          SDL_BlitSurface(players[p].ship[(int)(15-players[p].angle/(2*M_PI/15.0))],NULL,screen,&rect);
	if(players[p].shieldup) {
	  rect.x-=8;
	  rect.y-=8;
	  SDL_BlitSurface(players[p].shield,NULL,screen,&rect);
	}
      }
    }
  }
}

void draw_player_statusbars() {
  int p,p1,p2;
  SDL_Rect rect;
  rect.y=235;
  rect.w=280;
  rect.h=2;
  p=0;
  for(p1=0;p1<2;p1++) {
    rect.x=20;
    for(p2=0;p2<2;p2++) {
      if(players_active[p]) {
        rect.w=(double)280*players[p].health;
        SDL_FillRect(screen,&rect,plr_healthbar_col);
        if(players[p].energy<0)
          rect.w=0;
        else
          rect.w=(double)280*players[p].energy;
        rect.y+=2;
        SDL_FillRect(screen,&rect,plr_energybar_col);
        rect.y-=2;
        rect.x+=320;
      }
      p++;
    }
    rect.y+=240;
  }
}

void animate_player(int num) {
  Vector tmpv,sv;
  SDL_Rect *rect;
  Particle *part;
  char solid;
  int newx,newy,p;
  if(players[num].dead==0) {
    game_status.lifetime[num]++;
    players[num].angle+=players[num].turn*!players[num].frozen;
    if(players[num].visible==0) {
      players[num].energy-=0.001;
      if(players[num].energy<=0) players[num].visible=1;
    }
    if(players[num].ghost==1) {
      players[num].energy-=0.003;
      if(players[num].energy<=0) { players[num].ghost=0; ghostify(num,0); }
    }
    if(players[num].shieldup==1) {
      players[num].energy-=0.004;
      if(players[num].energy<=0) players[num].shieldup=0;
    }
    if(players[num].afterburn>1) {
      players[num].energy-=0.003;
      if(players[num].energy<=0) players[num].afterburn=1;
    }
    if(players[num].angle>2*M_PI) players[num].angle=0;
    if(players[num].angle<0) players[num].angle=2*M_PI;
    if(players[num].thrust || players[num].afterburn>1) {
      if(players[num].afterburn==1)
      tmpv = makeVector(sin(players[num].angle)*THRUST*players[num].thrust,cos(players[num].angle)*THRUST*players[num].thrust);
      else
      tmpv = makeVector(sin(players[num].angle)*THRUST*players[num].afterburn,cos(players[num].angle)*THRUST*players[num].afterburn);
      tmpv.y-=gravity.y;
      sv=addVectors(&players[num].vector,&tmpv);
      part=make_particle(players[num].x,players[num].y,15);
      part->vector=oppositeVector(&tmpv);
      if(hit_solid(players[num].x,players[num].y)<0) {
        part->color[0]=100; part->color[1]=100; part->color[2]=255;
        part->targ_color[0]=0; part->targ_color[1]=0; part->targ_color[2]=100;
        part->vector.x+=0.5-((rand()%100)/100.0);
        part->vector.y+=0.5-((rand()%100)/100.0);
        calc_color_deltas(part);
      }
      addParticle(part);
      if(players[num].afterburn>1) start_burning(players[num].x,players[num].y);
    } else sv=players[num].vector;
  } else sv=players[num].vector;
  tmpv=addVectors(&sv,&gravity);
  if(tmpv.x>MAXSPEED*players[num].afterburn) tmpv.x=MAXSPEED*players[num].afterburn;
  if(tmpv.y>MAXSPEED*players[num].afterburn) tmpv.y=MAXSPEED*players[num].afterburn;
  if(tmpv.x<-MAXSPEED*players[num].afterburn) tmpv.x=-MAXSPEED*players[num].afterburn;
  if(tmpv.y<-MAXSPEED*players[num].afterburn) tmpv.y=-MAXSPEED*players[num].afterburn;
  players[num].vector=tmpv;
  newx=players[num].x-(int)tmpv.x;
  newy=players[num].y-(int)tmpv.y;
  if(((newx<0 || newy<0) || (newx>lev_level.width-30 || newy>lev_level.height-30)) && players[num].dead==1) {
    players[num].dead=2;
    players[num].visible=0;
    plr_alive--;
    spawn_clusters(players[num].x,players[num].y,32,num);
  }
  if(newx<10) newx=10; if(newx>lev_level.width-16) newx=lev_level.width-16;
  if(newy<10) newy=10; if(newy>lev_level.height-16) newy=lev_level.height-16;
  solid=hit_solid(newx,newy);
  if(solid<0 && players[num].ghost==0) {
    if(players[num].dead==0)
      players[num].vector.y+=GRAVITY*2.1;
    else
      players[num].vector.y+=GRAVITY*0.5;
    if(players[num].afterburn>1) players[num].afterburn=1;
    if(solid==-2) { players[num].vector.y=MAXSPEED;  }
    else if(solid==-3) { players[num].vector.x=-MAXSPEED; }
    else if(solid==-4) { players[num].vector.y=-MAXSPEED; }
    else if(solid==-5) { players[num].vector.x=MAXSPEED; }
    else {
      if(players[num].vector.x>MAXSPEED/2) players[num].vector.x=MAXSPEED/2;
      if(players[num].vector.y>MAXSPEED/2) players[num].vector.y=MAXSPEED/2;
      if(players[num].vector.x<-MAXSPEED/2) players[num].vector.x=-MAXSPEED/2;
      if(players[num].vector.y<-MAXSPEED/2) players[num].vector.y=-MAXSPEED/2;
    }
  } else if(solid>0 && players[num].ghost==0) { 
    if(players[num].dead==1) {
      players[num].dead=2;
      players[num].visible=0;
      plr_alive--;
      spawn_clusters(players[num].x,players[num].y,32,num);
    }
    players[num].frozen=0;
    newx=players[num].x; newy=players[num].y;
    /* Collision damage */
    if(game_settings.coll_damage) 
      players[num].health-=fabs((double)((int)hypot(players[num].vector.x,players[num].vector.y)-1)/(MAXSPEED*24));
    if(players[num].health<=0 && !players[num].dead) { players[num].health=0; players[num].dead=1; }
    if(lev_level.solid[players[num].x][players[num].y]!=TER_INDESTRUCT) lev_level.solid[players[num].x][players[num].y]=0;
    players[num].vector.x=0;
    players[num].vector.y=0;
    /* Is the player sitting on a base ? */
    if(solid==TER_BASE) {
      if(players[num].health<1) players[num].health+=0.01;
      if(players[num].energy<1) players[num].energy+=0.03;
      if(players[num].angle>0 && players[num].angle<M_PI) players[num].angle-=0.1;
      if(players[num].angle>=M_PI) players[num].angle+=0.1;
    }
  } 
  /* Mid-air collisions */
  if(game_settings.ship_collisions && !players[num].ghost)
    for(p=0;p<4;p++) {
      if(num==p || players_active[p]==0) continue;
      if(abs(players[p].x-newx)<8 && abs(players[p].y-newy)<8) {
	tmpv=addVectors(&players[num].vector,&players[p].vector);
	players[p].vector=oppositeVector(&tmpv);
	break;
      }
    }

  players[num].x=newx;
  players[num].y=newy;
  switch(num) {
    case 0: rect=&cam1_rect; break;
    case 1: rect=&cam2_rect; break;
    case 2: rect=&cam3_rect; break;
    case 3: rect=&cam4_rect; break;
  }
  rect->x=newx-180;
  if(rect->x<0) rect->x=0;
  if(rect->x>lev_level.width-320) rect->x=lev_level.width-320;
  rect->y=newy-120;
  if(rect->y<0) rect->y=0;
  if(rect->y>lev_level.height-240) rect->y=lev_level.height-240;
  if(plr_alive<=1 && endgame==-1) endgame=30;
}

void player_keyhandler(SDL_KeyboardEvent *event,Uint8 type) {
  int p;
  SDLKey key=event->keysym.sym;
  for(p=0;p<4;p++) {
    if(players[p].frozen) continue;
    if(key==game_settings.buttons[p][0]) {
      if(type==SDL_KEYDOWN) players[p].thrust=1; else players[p].thrust=0;
    } else if(key==game_settings.buttons[p][1]) {
      if(type==SDL_KEYDOWN) players[p].thrust=-1; else players[p].thrust=0;
    } else if(key==game_settings.buttons[p][2]) {
      if(type==SDL_KEYDOWN) players[p].turn=TURN_SPEED; else players[p].turn=0;
    } else if(key==game_settings.buttons[p][3]) {
      if(type==SDL_KEYDOWN) players[p].turn=-TURN_SPEED; else players[p].turn=0;
    } else if(key==game_settings.buttons[p][4]) {
      if(type==SDL_KEYUP) fire_weapon(p,0);
    } else if(key==game_settings.buttons[p][5]) {
      if(type==SDL_KEYUP) fire_weapon(p,1);
    }
  }
}

void fire_weapon(int plr,int weapon) {
  Projectile *p;
  Vector v;
  if(players[plr].dead) return;
  /* First of all, check if we need any projectiles at all */
  if(weapon==1) {
    if(players[plr].special==WCloak) {
      if(players[plr].visible==0)
        players[plr].visible=1;
      else if(players[plr].energy>0)
        players[plr].visible=0;
      return;
    } else if(players[plr].special==WShield) {
      if(players[plr].shieldup==1)
        players[plr].shieldup=0;
      else if(players[plr].energy>0)
        players[plr].shieldup=1;
      return;
    } else if(players[plr].special==WGhostShip) {
      if(players[plr].ghost==1) {
        players[plr].ghost=0;
	ghostify(plr,0);
      } else if(players[plr].energy>0) {
        players[plr].ghost=1;
	ghostify(plr,1);
      }
      return;
    } else if(players[plr].special==WAfterBurner) {
      if(players[plr].afterburn>1) {
        players[plr].afterburn=1;
      } else if(players[plr].energy>0) {
	players[plr].afterburn=2;
      }
      return;
    } else if(players[plr].special==WWarp) {
      if(players[plr].energy>0) {
        drop_jumppoint(players[plr].x,players[plr].y,plr);
	players[plr].energy-=0.1;
      }
      return;
    }
  }
#ifdef HIRES_PROJ
  v=makeVector(sin(players[plr].angle),cos(players[plr].angle));
#else
  v=makeVector(sin(players[plr].angle)*1.6,cos(players[plr].angle)*1.6);
#endif
  p=make_projectile(players[plr].x,players[plr].y,v);
  p->owner=plr;
  if(weapon==1) { /* Special weapon */
    if(players[plr].energy<=0) {free(p); return;} else
    switch((int)players[plr].special) {
      case WGrenade:
        p->type=Grenade;
        players[plr].energy-=0.09;
	p->color=col_gray;
        break;
      case WMegaBomb:
        p->type=MegaBomb;
        p->vector.x=0;
        p->vector.y=0;
        p->gravity=&gravity;
        players[plr].energy-=0.2;
        break;
      case WMissile:
        p->type=Missile;
        p->angle=0;
        players[plr].energy-=0.18;
        break;
      case WMine:
	if(hit_solid(players[plr].x,players[plr].y)<0) {free(p); return;}
	p->type=Mine;
	p->vector.x=0;
	p->vector.y=0;
	p->owner=-1;
	p->primed=19;
	p->color=col_gray;
	players[plr].energy-=0.06;
	break;
      case WMagMine:
	if(hit_solid(players[plr].x,players[plr].y)<0) {free(p); return;}
	p->type=MagMine;
	p->vector.x=-p->vector.x/3;
	p->vector.y=-p->vector.y/3;
	p->color=SDL_MapRGB(screen->format,149,158,255);
	players[plr].energy-=0.06;
	p->owner=-1;
	p->primed=40;
	break;
      case WClaybomb:
	p->type=Claybomb;
	p->color=col_clay;
	players[plr].energy-=0.125;
	break;
      case WSnowball:
	p->type=Snowball;
	p->color=col_snow;
	players[plr].energy-=0.0833;
	break;
      case WPlastique:
	p->type=Plastique;
	p->color=col_gray;
	p->vector.x=-p->vector.x;
	p->vector.y=-p->vector.y;
	p->life=10;
	players[plr].energy-=0.01;
	break;
    }
  }
  add_projectile(p);

}

inline char hit_player(int x,int y,signed char ignore,char wide) {
  int p;
  for(p=0;p<4;p++)
    if(players_active[p] && p!=ignore && !players[p].dead)
      if(x>=players[p].x-wide && x<=players[p].x+wide)
        if(y>=players[p].y-wide && y<=players[p].y+wide) {
          return p+1;
        }
  return 0;
}

char player_damage(int plr,ProjectileType type) {
  double damage;
  switch(type) {
    case Missile:
      damage=0.02;
      break;
    case Grenade:
    case Mine:
    case MagMine:
      damage=0;
      break;
    case Snowball:
      players[plr].frozen=1;
      players[plr].thrust=0;
      players[plr].turn=0;
      return -1;
      break;
    default:
      damage=0.05;
  }
  players[plr].health-=damage;
  if(players[plr].health<=0) {
    players[plr].health=0;
    players[plr].dead=1;
  }
  return 1;
}

void ghostify(unsigned char plr,char status) {
  float alpha;
  int r;
  if(status==0) alpha=2; else alpha=0.5;
  for(r=0;r<SHIP_POSES;r++) {
    recolor(players[plr].ship[r],1,1,1,alpha);
  }
}
