#include <stdlib.h>
#include <math.h>
#include <SDL.h>
#include "defines.h"
#include "fs.h"
#include "player.h"
#include "level.h"
#include "console.h"
#include "special.h"

struct Special_list {
  SpecialObj *special;
  struct Special_list *next;
  struct Special_list *prev;
};
/* Globals */
struct Special_list *specials;

SDL_Surface *fx_jumpgate[1];
SDL_Surface *fx_jumppoint_entry[WARP_FRAMES];
SDL_Surface *fx_jumppoint_exit[WARP_FRAMES];

/* Internally used function */
char hitsolid_rect(int x,int y,int w,int h) {
  int tx,ty,tx2,ty2;
  tx2=x+w; if(tx2>lev_level.width) return 1;
  ty2=y+h; if(ty2>lev_level.height) return 1;
  for(tx=x;tx<tx2;tx++)
    for(ty=y;ty<ty2;ty++)
      if(lev_level.solid[tx][ty]!=TER_FREE&&lev_level.solid[tx][ty]!=TER_WATER) return 1;
  return 0;
}

void init_specials() {
  char filename[11];
  int r;
  specials=NULL;
  for(r=0;r<WARP_FRAMES;r++) {
    sprintf(filename,"warp%.2d.tga",r); 
    fx_jumppoint_entry[r]=load_image(GFX_DIRECTORY,filename,0,1);
    sprintf(filename,"warpe%.2d.tga",r);
    fx_jumppoint_exit[r]=load_image(GFX_DIRECTORY,filename,0,1);
  }
  fx_jumpgate[0]=load_image(GFX_DIRECTORY,"jumpgate.tga",0,1);
}

void deinit_specials(char just_clear) {
  int r;
  struct Special_list *next;
  if(just_clear==0)
    for(r=0;r<WARP_FRAMES;r++)
      SDL_FreeSurface(fx_jumppoint_entry[r]);
  while(specials) {
    next=specials->next;
    free(specials->special);
    free(specials);
    specials=next;
  }
}

void prepare_specials() {
  SpecialObj *gate1,*gate2;
  int r;
  for(r=0;r<game_settings.jumpgates;r++) {
    gate1=(SpecialObj*)malloc(sizeof(SpecialObj));
    gate1->age=-1;
    gate1->frame=0;
    gate1->type=JumpGate;
    gate1->owner=-1;
    gate1->timer=0;
    gate1->var1=0;
    do {
      gate1->x=rand()%lev_level.width;
      gate1->y=rand()%lev_level.height;
    } while(hitsolid_rect(gate1->x,gate1->y,32,32));
    gate2=(SpecialObj*)malloc(sizeof(SpecialObj));
    gate2->age=-1;
    gate2->frame=0;
    gate2->type=JumpGate;
    gate2->owner=-1;
    gate2->timer=0;
    gate2->var1=0;
    do {
      gate2->x=rand()%lev_level.width;
      gate2->y=rand()%lev_level.height;
    } while(hitsolid_rect(gate2->x,gate2->y,32,32) || hypot(gate2->x-gate1->x,gate2->y-gate1->y)<200);
    gate1->link=gate2;
    gate2->link=gate1;
    addspecial(gate2);
    addspecial(gate1);
  }
}

void addspecial(SpecialObj *special) {
  struct Special_list *list=NULL;
  struct Special_list *newentry;
  if(specials!=NULL) {
    list=specials;
    while(list->next) list=list->next;
  }
  newentry=(struct Special_list*)malloc(sizeof(struct Special_list));
  newentry->next=NULL;
  newentry->prev=list;
  newentry->special=special;
  if(specials==NULL)
    specials=newentry;
  else
    list->next=newentry;
}

void drop_jumppoint(int x,int y,char player) {
  struct Special_list *list=specials;
  SpecialObj *jump,*exit=NULL;
  jump=(SpecialObj*)malloc(sizeof(SpecialObj));
  jump->x=x-16;
  jump->y=y-16;
  jump->owner=player;
  jump->frame=0;
  jump->timer=0;
  while(list) {
    if(list->special->type==WarpExit && list->special->owner==player)
      if(list->special->link==NULL) exit=list->special;
    list=list->next;
  }
  if(exit==NULL) {
    jump->type=WarpExit;
    jump->link=NULL;
    jump->age=-1;
  } else {
    jump->type=WarpEntry;
    jump->link=exit;
    exit->link=jump;
    jump->age=(game_settings.jumplife)?JPLONGLIFE:JPSHORTLIFE;
    exit->age=(game_settings.jumplife)?JPLONGLIFE:JPSHORTLIFE;
  }
  addspecial(jump);
}

void animate_specials() {
  struct Special_list *list=specials;
  struct Special_list *next;
  while(list) {
    next=list->next;
    /* Animations */
    switch(list->special->type) {
      case Unknown:
      case JumpGate: break;
      case WarpExit:
      case WarpEntry:
	list->special->frame++;
	if(list->special->age<0) list->special->frame=0;
	else if(list->special->frame>=WARP_FRAMES) list->special->frame=WARP_LOOP;
	break;
    }
    /* Decrement age if positive */
    if(list->special->age>0) list->special->age--;
    if(list->special->timer>0) list->special->timer--;
    /* Check if the object has expired. If age is <0, the object has no time limit */
    if(list->special->age==0) {
      free(list->special);
      if(list->prev)
        list->prev->next=list->next;
      else
        specials=list->next;
      if(list->next)
        list->next->prev=list->prev;
      free(list);
    }
    /* Move on to the next item */
    list=next;
  }
}

void draw_specials() {
  struct Special_list *list=specials;
  SDL_Surface **sprite;
  SDL_Rect rect;
  signed char showme=-1;
  while(list) {
    switch(list->special->type) {
      case Unknown: sprite=NULL; break;
      case WarpEntry: sprite=fx_jumppoint_entry; if(list->special->frame==0) showme=list->special->owner; break;
      case WarpExit: sprite=fx_jumppoint_exit; if(list->special->frame==0) showme=list->special->owner; break;
      case JumpGate: sprite=fx_jumpgate;
    }
    if(players_active[0] && (showme==0||showme==-1)) {
      rect.x=list->special->x-cam1_rect.x;
      rect.y=list->special->y-cam1_rect.y;
      if((rect.x>0 && rect.x<320) && (rect.y>0 && rect.y<240)) 
        SDL_BlitSurface(sprite[list->special->frame],NULL,screen,&rect);
    }
    if(players_active[1] && (showme==1||showme==-1)) {
      rect.x=list->special->x-cam2_rect.x+320;
      rect.y=list->special->y-cam2_rect.y;
      if((rect.x>320 && rect.x<640) && (rect.y>0 && rect.y<240)) 
	SDL_BlitSurface(sprite[list->special->frame],NULL,screen,&rect);
    }
    if(players_active[2] && (showme==2||showme==-1)) {
      rect.x=list->special->x-cam3_rect.x;
      rect.y=list->special->y-cam3_rect.y+240;
      if((rect.x>0 && rect.x<320) && (rect.y>240 && rect.y<480)) 
        SDL_BlitSurface(sprite[list->special->frame],NULL,screen,&rect);
    }
    if(players_active[3] && (showme==3||showme==-1)) {
      rect.x=list->special->x-cam4_rect.x+320;
      rect.y=list->special->y-cam4_rect.y+240;
      if((rect.x>320 && rect.x<640) && (rect.y>240 && rect.y<480))  
	SDL_BlitSurface(sprite[list->special->frame],NULL,screen,&rect);
    }
    list=list->next;
  }
}

void player_hit_specials() {
  struct Special_list *list=specials;
  int p,w;
  while(list) {
    w=32;
    for(p=0;p<4;p++) {
      if(players[p].x>=list->special->x && players[p].x<=list->special->x+w)
	if(players[p].y>=list->special->y && players[p].y<=list->special->y+w) {
	  switch(list->special->type) {
	    case WarpEntry:
	    case WarpExit:
	      if(list->special->link && list->special->timer==0) {
	        players[p].x=list->special->link->x+w/2;
	        players[p].y=list->special->link->y+w/2;
	        list->special->timer=25;
	        list->special->link->timer=25;
	      } break;
	    case JumpGate:
	      if(list->special->timer==0) {
		list->special->timer=(game_settings.jumplife)?JPLONGLIFE:JPSHORTLIFE;
		list->special->link->timer=(game_settings.jumplife)?JPLONGLIFE:JPSHORTLIFE;
	        drop_jumppoint(list->special->x+16,list->special->y+16,p+10);
		drop_jumppoint(list->special->link->x+16,list->special->link->y+16,p+10);
	      }
	      break;
	    default: break;
	  }
	}
    }
    list=list->next;
  }
}
