#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "defines.h"
#include "fs.h"
#include "console.h"
#include "level.h"
#include "weapon.h"
#include "player.h"
#include "particle.h"
#include "game.h"

/* Linked list containing all projectiles */
struct Projectile_list {
  Projectile *projectile;
  struct Projectile_list *next;
  struct Projectile_list *prev;
};
/* Linked list containing all explosions */
struct Explosion_list {
  Explosion *explosion;
  struct Explosion_list *next;
  struct Explosion_list *prev;
};

/* Internally used globals */
struct Projectile_list *projectiles;
struct Explosion_list *explosions;
Vector wea_gravity;
SDL_Surface *wea_explosion[EXPL_FRAMES];
SDL_Surface *wea_hole;
Uint8 *wea_gnaw;

/* Initialize */
void init_weapons() {
  int r,n,d;
  Uint8 *bits;
  char filename[]="expl00.tga";
  projectiles=NULL;
  wea_gravity.y=-WEAP_GRAVITY;
  wea_gravity.x=0;
  for(r=0;r<EXPL_FRAMES;r++) {
    memset(filename+4,'0',2);
    n=r+2;
    for(d=2;d>0;d--) {
      for(;n>pow(10,d-1);)
        if(n-=pow(10,d-1)) filename[6-d]++; else break;
    }
     wea_explosion[r]=load_image(GFX_DIRECTORY,filename,0,1);
  }
  wea_hole=load_image(GFX_DIRECTORY,"hole.png",0,1);
  wea_gnaw=(Uint8*)malloc(sizeof(Uint8)*wea_hole->w*wea_hole->h);
  if(wea_hole->format->BytesPerPixel!=4) {
    printf("Screen has %d bytes per pixel ! Currently only 4 bps supported :-( (Fix this at init_weapons(), weapons.c)\n",wea_hole->format->BytesPerPixel);
    exit(1);
  }

  for(r=0;r<wea_hole->w*wea_hole->h;r++) {
    bits=((Uint8*)wea_hole->pixels+r*4+3);
    wea_gnaw[r]=(*((Uint8*)(bits))!=0)?0:1;
  }
}

void deinit_weapons(char just_clear) {
  struct Explosion_list *enext;
  struct Projectile_list *pnext;
  int r;
  while(explosions) {
    enext=explosions->next;
    free(explosions->explosion);
    free(explosions);
    explosions=enext;
  }
  while(projectiles) {
    pnext=projectiles->next;
    free(projectiles->projectile);
    free(projectiles);
    projectiles=pnext;
  }
  if(!just_clear) {
    SDL_FreeSurface(wea_hole);
    for(r=0;r<EXPL_FRAMES;r++)
      SDL_FreeSurface(wea_explosion[r]);
    free(wea_gnaw);
  }
}
Projectile *make_projectile(int x,int y,Vector v) {
  Projectile *newproj;
  newproj=(Projectile*)malloc(sizeof(Projectile));
  newproj->x=x;
  newproj->y=y;
  newproj->vector=v;
  newproj->gravity=NULL;
  newproj->vector.x=newproj->vector.x*5;
  newproj->vector.y=newproj->vector.y*5;
  newproj->color=SDL_MapRGB(screen->format,255,100,100);
  newproj->type=Cannon;
  newproj->owner=-1;
  newproj->primed=0;
  newproj->life=-1;
  return newproj;
}
/* Add a new projectile to list */
void add_projectile(Projectile *proj) {
  struct Projectile_list *list=NULL;
  struct Projectile_list *newentry;
  if(projectiles!=NULL) {
    list=projectiles;
    while(list->next) list=list->next;
  }
  newentry=(struct Projectile_list*)malloc(sizeof(struct Projectile_list));
  newentry->next=NULL;
  newentry->prev=list;
  newentry->projectile=proj;
  if(projectiles==NULL)
    projectiles=newentry;
  else
    list->next=newentry;
}

void add_explosion(int x,int y) {
  struct Explosion_list *list=NULL;
  struct Explosion_list *newentry;
  char solid;
  int fx,fy,rx,ry;
  if(explosions!=NULL) {
    list=explosions;
    while(list->next) list=list->next;
  }
  newentry=(struct Explosion_list*)malloc(sizeof(struct Explosion_list));
  newentry->next=NULL;
  newentry->prev=list;
  newentry->explosion=(Explosion*)malloc(sizeof(Explosion));
  newentry->explosion->x=x;
  newentry->explosion->y=y;
  newentry->explosion->frame=0;
  if(explosions==NULL)
    explosions=newentry;
  else
    list->next=newentry;
  /* Gnaw a hole in terrain */
  for(fx=0;fx<wea_hole->w;fx++)
    for(fy=0;fy<wea_hole->h;fy++) {
      rx=fx-wea_hole->w/2+x;
      ry=fy-wea_hole->h/2+y;
      if(rx<0 || ry<0 || rx>=lev_level.width || ry>=lev_level.height) continue;
      if(lev_level.solid[rx][ry]==TER_INDESTRUCT || hit_solid(rx,ry)<0) continue;
      if(game_settings.indstr_base && lev_level.solid[rx][ry]==TER_BASE) continue;
      solid=lev_level.solid[rx][ry];
      if(solid==TER_FREE) continue;
      lev_level.solid[rx][ry]*=wea_gnaw[fy*wea_hole->w+fx];
      if(!lev_level.solid[rx][ry] && solid==TER_UNDERWATER) lev_level.solid[rx][ry]=TER_WATER;
      solid=hit_solid(rx,ry);
      if(solid<=0) putpixel(lev_level.terrain,rx,ry,(solid<0)?lev_watercol:0);
    }
}

void animate_weapons() {
  struct Projectile_list *plist=projectiles,*pnext;
  struct Explosion_list *elist=explosions,*enext;
  signed char explode,solid;
  int p,i1,i2;
  double f1,f0;
#ifdef HIRES_PROJ
  int h1;
#endif
  Projectile *proj;
  Explosion *expl;
  if(plist) {
    while(plist) {
      proj=plist->projectile;
      pnext=plist->next;

      /* Weapon specific movement code */
      if(proj->type==Missile) {
        p=find_nearest_player(proj->x,proj->y,proj->owner,&f0);
        i1=players[p].x-proj->x;
        i2=players[p].y-proj->y;
        f1=atan2(i1,i2);
        if(proj->angle<f1) proj->angle+=0.1; else proj->angle-=0.1;
	if(proj->angle<-M_PI) proj->angle=M_PI;
	else if(proj->angle>M_PI) proj->angle=-M_PI;
	if(players[p].shieldup && f0 < 70) { if(proj->angle>f1) proj->angle+=1; else proj->angle-=1; }
        proj->vector=makeVector(sin(proj->angle)*-6,cos(proj->angle)*-6);
        addParticle(make_particle(proj->x,proj->y,5));
      } else if(proj->type==MagMine && proj->primed<=0) {
        p=find_nearest_player(proj->x,proj->y,proj->owner,&f1);
	if(f1<70) {
	  proj->vector.x=(-(players[p].x-proj->x)/4.0);
	  proj->vector.y=(-(players[p].y-proj->y)/4.0);
	} else { proj->vector.x=0; proj->vector.y=0; }
      } else if(proj->type==Mine && proj->primed<=0) {
        p=find_nearest_player(proj->x,proj->y,-1,&f1);
	if(f1<40) proj->color=col_gray; else proj->color=0;
      }
      /* Check if someone is operating a force field nearby */
      for(p=0;p<4;p++) {
        if(players[p].shieldup) {
	  f1=hypot(abs(players[p].x-proj->x),abs(players[p].y-proj->y));
	  if(f1<100) {
	    proj->vector.x+=((players[p].x-proj->x)/(abs(players[p].x-proj->x)+0.1))/2.0;
	    proj->vector.y+=((players[p].y-proj->y)/(abs(players[p].y-proj->y)+0.1))/2.0;
	  }
	}
      }
      if(proj->life>0) proj->life--;
#ifdef HIRES_PROJ
      explode=0;
      for(h1=0;h1<2;h1++) {
        if(explode) break;
#endif
      proj->x-=proj->vector.x;
      proj->y-=proj->vector.y;
      if(proj->x<0) proj->x=0;
      if(proj->y<0) proj->y=0;
      if(proj->gravity)
        proj->vector=addVectors(&proj->vector,proj->gravity);
      if(proj->primed<=0) explode=hit_player(proj->x,proj->y,proj->owner,(proj->type==Mine||proj->type==MagMine||proj->type==Grenade||proj->type==Missile)?16:8); else {proj->primed--; explode=0;}
      if(explode) explode=player_damage(explode-1,proj->type);
      if(proj->x<=wea_hole->w) explode++;  if(proj->x>=lev_level.width-wea_hole->w) explode++;
      if(proj->y<=wea_hole->h) explode++;  if(proj->y>=lev_level.height-wea_hole->h) explode++;
      solid=hit_solid(proj->x,proj->y);
      if(solid<0) {
	if(proj->type==Snowball) explode++;
        if(proj->vector.x>1.3) proj->vector.x=1.3;
	else if(proj->vector.x<-1.3) proj->vector.x=-1.3;
	if(proj->vector.y>1.3) proj->vector.y=1.3;
	else if(proj->vector.y<-1.3) proj->vector.y=-1.3;
      }
      explode+=(solid>0);
      if(explode) {
	if(lev_level.solid[proj->x][proj->y]==TER_EXPLOSIVE) spawn_clusters(proj->x,proj->y,6,-1);
	else if(lev_level.solid[proj->x][proj->y]==TER_COMBUSTABLE) {
	  start_burning(proj->x+10,proj->y);
	  start_burning(proj->x-10,proj->y);
	  start_burning(proj->x,proj->y+11);
	  start_burning(proj->x,proj->y-11);

	}
	if(proj->type==Snowball) { if(explode>0) alter_level(proj->x,proj->y,15,Ice); }
	else if(proj->type==Claybomb) alter_level(proj->x,proj->y,15,Earth);
	else if(proj->type==Plastique) alter_level(proj->x,proj->y,10,Explosive);
	else add_explosion(proj->x,proj->y);
        switch(proj->type) {
          case Grenade: spawn_clusters(proj->x,proj->y,16,-1); break;
          case MegaBomb: spawn_clusters(proj->x,proj->y,32,-1); break;
	  case Missile: spawn_clusters(proj->x,proj->y,10,-1); break;
	  case MagMine:
	  case Mine:
	    spawn_clusters(proj->x,proj->y,8,-1); break;
          default: break;
        }
        free(proj);
        if(plist->prev)
          plist->prev->next = plist->next;
        else
          projectiles=plist->next;
        if(plist->next)
          plist->next->prev = plist->prev;
        free(plist); plist=NULL;
      }
#ifdef HIRES_PROJ
      }
#endif
      if(plist) {
      if(proj->life==0) {
        free(proj);
	if(plist->prev)
	  plist->prev->next = plist->next;
	else
	  projectiles=plist->next;
	if(plist->next)
	  plist->next->prev = plist->prev;
	free(plist);
      }
      }
      plist=pnext;
    }
  }
  if(elist) {	/* Explosions */
    while(elist) {
      expl=elist->explosion;
      enext=elist->next;
      expl->frame++;
      if(expl->frame==EXPL_FRAMES) { /* Animation is over */
        free(expl);
        if(elist->prev)
          elist->prev->next = elist->next;
        else
          explosions=elist->next;
        if(elist->next)
          elist->next->prev = elist->prev;
        free(elist);
      }
      elist=enext;
    }
  }
}

inline void draw_special_weapons(int x,int y,ProjectileType type,Uint32 color) {
switch(type) {
  case MegaBomb:
    putpixel(screen,x,y-1,color);
    putpixel(screen,x,y+1,color);
    putpixel(screen,x-1,y-1,color);
    putpixel(screen,x+1,y-1,color);
    break;
  case Mine:
  case MagMine:
  case Missile:
    if(color!=0) {
    putpixel(screen,x,y-1,color);
    putpixel(screen,x,y+1,color);
    putpixel(screen,x-1,y,color);
    putpixel(screen,x+1,y,color);
    }
  
  default: break;
}
}

void draw_weapons() {
  struct Projectile_list *plist=projectiles;
  struct Explosion_list *elist=explosions;
  SDL_Surface *tmpsurf;
  int x,y;
  SDL_Rect rect;
  if(plist==NULL && elist==NULL) return;
  if( SDL_MUSTLOCK(screen) ) SDL_LockSurface(screen);
  /* Draw projectiles */
  while(plist) {
    if(players_active[0]) {
      x=plist->projectile->x-cam1_rect.x;
      y=plist->projectile->y-cam1_rect.y;
      if((x>0 && x<320) && (y>0 && y<240)) {
        putpixel(screen,x,y,plist->projectile->color);
	draw_special_weapons(x,y,plist->projectile->type,plist->projectile->color);
      }
    }
    if(players_active[1]) {
      x=plist->projectile->x-cam2_rect.x;
      y=plist->projectile->y-cam2_rect.y;
      if((x>0 && x<320) && (y>0 && y<240)) {
        putpixel(screen,320+x,y,plist->projectile->color);
	draw_special_weapons(320+x,y,plist->projectile->type,plist->projectile->color);
      }
    }
    if(players_active[2]) {
      x=plist->projectile->x-cam3_rect.x;
      y=plist->projectile->y-cam3_rect.y;
      if((x>0 && x<320) && (y>0 && y<240)) {
        putpixel(screen,x,240+y,plist->projectile->color);
	draw_special_weapons(x,240+y,plist->projectile->type,plist->projectile->color);
      }
    }
    if(players_active[3]) {
      x=plist->projectile->x-cam4_rect.x;
      y=plist->projectile->y-cam4_rect.y;
      if((x>0 && x<320) && (y>0 && y<240)) {
        putpixel(screen,320+x,240+y,plist->projectile->color);
	draw_special_weapons(320+x,240+y,plist->projectile->type,plist->projectile->color);
      }
    }
    plist=plist->next;
  }
  if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);
  while(elist) {
    tmpsurf=wea_explosion[elist->explosion->frame];
    if(players_active[0]) {
      rect.x=elist->explosion->x-cam1_rect.x-8;
      rect.y=elist->explosion->y-cam1_rect.y-8;
      if((rect.x>0 && rect.x<320) && (rect.y>0 && rect.y<240))
        SDL_BlitSurface(tmpsurf,NULL,screen,&rect);
    }
    if(players_active[1]) {
      rect.x=320+elist->explosion->x-cam2_rect.x-8;
      rect.y=elist->explosion->y-cam2_rect.y-8;
      if((rect.x>320 && rect.x<640) && (rect.y>0 && rect.y<240))
        SDL_BlitSurface(tmpsurf,NULL,screen,&rect);
    }
    if(players_active[2]) {
      rect.x=elist->explosion->x-cam3_rect.x-8;
      rect.y=240+elist->explosion->y-cam3_rect.y-8;
      if((rect.x>0 && rect.x<320) && (rect.y>240 && rect.y<480))
        SDL_BlitSurface(tmpsurf,NULL,screen,&rect);
    }
    if(players_active[3]) {
      rect.x=320+elist->explosion->x-cam4_rect.x-8;
      rect.y=240+elist->explosion->y-cam4_rect.y-8;
      if((rect.x>320 && rect.x<640) && (rect.y>240 && rect.y<480))
        SDL_BlitSurface(tmpsurf,NULL,screen,&rect);
    }

    elist=elist->next;
  }
}

void spawn_clusters(int x,int y,int count,int owner) {
  Projectile *cluster;
  double angle,incr;
  angle=0;
  incr=(2*M_PI)/(double)count;
  while(angle<2*M_PI) {
    cluster=make_projectile(x,y,makeVector(sin(angle),cos(angle)));
    cluster->type=Cannon;
    cluster->gravity=&wea_gravity;
    cluster->owner=owner;
    add_projectile(cluster);
    angle+=incr;
  }
}

int find_nearest_player(int myX,int myY,int not_this,double *dist) {
  int p,plr=-1;
  double distance=9999999,d;
  for(p=0;p<4;p++) {
    if(p==not_this || !players_active[p]) continue;
    d=hypot(abs(players[p].x-myX),abs(players[p].y-myY));
    if(d<distance) {plr=p; distance=d;}
  }
  if(dist) *dist=distance;
  return plr;
}
